package org.gersteinlab.act;

import org.gersteinlab.act.data.*;
import org.gersteinlab.act.io.*;
import org.gersteinlab.act.util.*;

import java.io.*;
import java.util.*;

/**
 *	<pre>
 *	This class converts an input signal file into a bin file with a
 *	specified bin size. The total length of each chromosome is specified in
 *	the configuration file.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (May 8, 2010)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class BinFileCreator
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	/**
	 *	Use the mean signal as the summary statistic of a bin.
	 */
	public static final String BIN_MEAN		= "Mean";

	/**
	 *	Use the maximum signal as the summary statistic of a bin.
	 */
	public static final String BIN_MAX		= "Max";

	/**
	 *	Use the presence of signal as the summary statistic of a bin.
	 */
	public static final String BIN_BINARY		= "Binary";

	/**
	 *	Use the mean signal as the summary statistic of a bin.
	 */
	protected static final int BIN_MEAN_ID		= 1;

	/**
	 *	Use the maximum signal as the summary statistic of a bin.
	 */
	protected static final int BIN_MAX_ID		= 2;

	/**
	 *	Use the presence of signal as the summary statistic of a bin.
	 */
	protected static final int BIN_BINARY_ID	= 3;


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	public static void main(String[] argv) throws Exception
	{
		// Check the arguments, and print usage if necessary
		if (argv.length != 5 ||
		    (!argv[1].equals(BIN_MEAN) && !argv[1].equals(BIN_MAX) && !argv[1].equals(BIN_BINARY)))
		{
			System.err.println("Usage: java -classpath <classpath> org.gersteinlab.act.BinFileCreator <Config file> <Bin statistic: {" + BIN_MEAN + " | " + BIN_MAX + " | " + BIN_BINARY + "}> <Bin size> <Output file> <Input file>");
			System.exit(-1);
		}

		// Register input parameters
		int argc = 0;
		Config config	= new Config(new File(argv[argc++]));
		String binStat	= argv[argc++];
		int binStatId	= -1;
		if (binStat.equals(BIN_MEAN))	binStatId = BIN_MEAN_ID;
		if (binStat.equals(BIN_MAX))	binStatId = BIN_MAX_ID;
		if (binStat.equals(BIN_BINARY))	binStatId = BIN_BINARY_ID;
		int binSize	= Integer.parseInt(argv[argc++]);
		if (binSize <= 0)
			throw new RuntimeException("Invalid bin size: " + binSize);
		File outFile	= new File(argv[argc++]);
		File inFile	= new File(argv[argc++]);

		// Get the relevant chromosomes
		Map<String, Integer> chromosomeIndices = new HashMap<String, Integer>();
		int index = 0;
		for (String configName : config.getConfigNames())
			if (configName.startsWith("length."))
				chromosomeIndices.put(configName.substring("length.".length()), new Integer(index++));

		// Get the chromosome lengths, and determine the number of bins required
		int chromosomeCount = chromosomeIndices.size();
		long[] chromosomeLengths = new long[chromosomeCount];
		int[] binCounts = new int[chromosomeCount];
		double[][] binStats = new double[chromosomeCount][];	// Use double for all statistics for simplicity
		index = 0;
		for (String configName : config.getConfigNames())
		{
			if (configName.startsWith("length."))
			{
				chromosomeLengths[index] = Long.parseLong(config.getConfig(configName));
				binCounts[index] = (int)(chromosomeLengths[index] / binSize);
				if (((long)binCounts[index]) * binSize != chromosomeLengths[index])
					binCounts[index] += 1;
				binStats[index] = new double[binCounts[index]];
				Arrays.fill(binStats[index], Double.NaN);
				index ++;
			}
		}

		// Read input file
		DataFilesReader dfr = DataFilesReader.getReader(config, inFile);
		Interval interval = dfr.readInterval();
		while (interval != null)
		{
			// Determine the bins involved
			int chrIndex = chromosomeIndices.get(interval.getChr());
			int binIndex = (int)interval.getStart() / binSize;
			long binStart = ((long)binIndex) * binSize;
			while (binStart < interval.getEnd())
			{
				// Got an error like this, therefore adding this check
				if (binIndex >= binCounts[chrIndex])
				{
					System.err.println("Warning: interval " + interval.toString() + " exceeds length of chromosome (" + chromosomeLengths[chrIndex] + ")");
					break;
				}

				     if (binStatId == BIN_MEAN_ID)
				{
					long binEnd = binStart + binSize;
					if (binEnd > chromosomeLengths[chrIndex])
						binEnd = chromosomeLengths[chrIndex];
					double partialSum = interval.getDoubleValue() * ((interval.getEnd() > binEnd ?binEnd :interval.getEnd()) - (interval.getStart() > binStart ?interval.getStart() :binStart));
					if (Double.isNaN(binStats[chrIndex][binIndex]))
						binStats[chrIndex][binIndex] = partialSum;
					else
						binStats[chrIndex][binIndex] += partialSum;
				}
				else if (binStatId == BIN_MAX_ID)
				{
					if (Double.isNaN(binStats[chrIndex][binIndex]) ||
					    interval.getDoubleValue() > binStats[chrIndex][binIndex])
						binStats[chrIndex][binIndex] = interval.getDoubleValue();
				}
				else if (binStatId == BIN_BINARY_ID)
					binStats[chrIndex][binIndex] = 1;

				binIndex ++;
				binStart += binSize;
			}

			interval = dfr.readInterval();
		}
		dfr.close();

		// Output the bins
		PrintWriter pw = new PrintWriter(new FileWriter(outFile));
		Iterator chrIter = chromosomeIndices.keySet().iterator();
		for (int i=0; i<chromosomeCount; i++)
		{
			pw.println("fixedStep chrom=" + chrIter.next() + " start=1 step=" + binSize + " span=" + binSize);
			int binStart = 0;
			for (int j=0; j<binCounts[i]; j++)
			{
				long binEnd = binStart + binSize;
				if (binEnd > chromosomeLengths[i])
					binEnd = chromosomeLengths[i];

				double outValue = binStats[i][j];
				if (Double.isNaN(outValue))	// This is what most people want
					outValue = 0;

				     if (binStatId == BIN_MEAN_ID)
					pw.println(outValue / (binEnd - binStart));
				else if (binStatId == BIN_MAX_ID)
					pw.println(outValue);
				else if (binStatId == BIN_BINARY_ID)
					pw.println((int)outValue);
				binStart += binSize;
			}
		}
		pw.close();
	}
}
