package org.gersteinlab.act;

import org.gersteinlab.act.data.*;
import org.gersteinlab.act.io.*;
import org.gersteinlab.act.util.*;

import JSci.maths.statistics.*;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

/**
 *	<pre>
 *	This class computes the correlations between different datasets. It
 *	takes a number of data files (in WIG or BED format) as input, and
 *	computes the correlation between each pair of files.
 *
 *	The correlation function can be specified as Pearson, Spearman or
 *	normal-score.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (May 7, 2010)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class CorrelationsCalculator
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	/**
	 *	Pearson correlation
	 */
	public static final String PEARSON = "Pearson";

	/**
	 *	Spearman correlation
	 */
	public static final String SPEARMAN = "Spearman";

	/**
	 *	Normal-score correlation
	 */
	public static final String NORMALSCORE = "NormalScore";


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	public static void main(String[] argv) throws Exception
	{
		// Check the arguments, and print usage if necessary
		if (argv.length < 5 ||
		    (!argv[1].equals(PEARSON) && !argv[1].equals(SPEARMAN) && !argv[1].equals(NORMALSCORE)))
		{
			System.err.println("Usage: java -classpath <classpath> org.gersteinlab.act.CorrelationsCalculator <Config file> <Correlation function: {" + PEARSON + " | " + SPEARMAN + " | " + NORMALSCORE + "}> <Output file> <Input file 1> <Input file 2> ...");
			System.exit(-1);
		}

		// Register input parameters
		int argc = 0;
		Config config	= new Config(new File(argv[argc++]));
		String corrFunc	= argv[argc++];
		File outFile	= new File(argv[argc++]);

		int inFileCount = argv.length - argc;
		File[] inFiles = new File[inFileCount];
		for (int i=0; i<inFileCount; i++)
			inFiles[i] = new File(argv[argc++]);

		// Compute correlations
		double[][] correlations = new double[inFileCount][];
		for (int i=0; i<inFileCount; i++)
			correlations[i] = new double[i];

		for (int i=inFileCount-1; i>=0; i--)
		{
			double[] vi = null;	// Values
			double[] ri = null;	// Ranks
			double[] zi = null;	// Normal scores

			vi = readVector(inFiles[i], config);
			if (corrFunc.equals(SPEARMAN) || corrFunc.equals(NORMALSCORE))
			{
				ri = getRanks(vi);
				vi = null;

				if (corrFunc.equals(NORMALSCORE))
				{
					zi = getNormalScores(ri);
					ri = null;
				}
			}

			for (int j=0; j<i; j++)
			{
				double[] vj = null;	// Values
				double[] rj = null;	// Ranks
				double[] zj = null;	// Normal scores

				vj = readVector(inFiles[j], config);
				if (corrFunc.equals(SPEARMAN) || corrFunc.equals(NORMALSCORE))
				{
					rj = getRanks(vj);
					vj = null;

					if (corrFunc.equals(NORMALSCORE))
					{
						zj = getNormalScores(rj);
						rj = null;
					}
				}

				     if (corrFunc.equals(PEARSON))
					correlations[i][j] = getPearsonCorrelation(vi, vj);
				else if (corrFunc.equals(SPEARMAN))
					correlations[i][j] = getPearsonCorrelation(ri, rj);
				else if (corrFunc.equals(NORMALSCORE))
					correlations[i][j] = getPearsonCorrelation(zi, zj);
			}
		}

		// Output correlations
		PrintWriter pw = new PrintWriter(new FileWriter(outFile));
		     if (corrFunc.equals(PEARSON))
			pw.print("Pearson Correlation");
		else if (corrFunc.equals(SPEARMAN))
			pw.print("Spearman Correlation");
		else if (corrFunc.equals(NORMALSCORE))
			pw.print("Normal score Correlation");

		for (int j=0; j<inFileCount; j++)
			pw.print("\t" + inFiles[j].getName());
		pw.println();

		for (int i=0; i<inFileCount; i++)
		{
			pw.print(inFiles[i].getName());
			for (int j=0; j<inFileCount; j++)
			{
				pw.print("\t");
				     if (j < i)
					pw.print(correlations[i][j]);
				else if (i == j)
					pw.print("1");
				else
					pw.print(correlations[j][i]);
			}
			pw.println();
		}
		pw.close();
	}

	/**
	 *	Read a vector of values from a data file.
	 *	@param		inFile			The data file
	 *	@param		config			For getting file format configurations
	 *	@return					The vector
	 */
	public static double[] readVector(File inFile, Config config)
	throws ClassNotFoundException,
	       IllegalAccessException,
	       InstantiationException,
	       InvocationTargetException,
	       IOException,
	       NoSuchMethodException
	{
		// Count the number of intervals
		DataFilesReader dfr = DataFilesReader.getReader(config, inFile);
		int intervalCount = dfr.countIntervals();
		dfr.close();

		// Read the values
		double[] values = new double[intervalCount];
		dfr = DataFilesReader.getReader(config, inFile);
		for (int i=0; i<intervalCount; i++)
			values[i] = dfr.readInterval().getDoubleValue();
		dfr.close();
		return values;
	}

	/**
	 *	Get the unique values and their ranks [0..length-1] in a vector.
	 *	If a value appears multiple times, its rank will be the middle
	 *	location of all appearences.
	 *	@param		v			The values
	 *	@return					A 2-D array of values and ranks
	 */
	public static double[][] getUniqueValues(double[] v)
	{
		if (v == null)
			return null;

		if (v.length == 0)
			return new double[2][0];

		// Sort the values
		double[] v2 = new double[v.length];
		for (int i=0; i<v.length; i++)
			v2[i] = v[i];
		Arrays.sort(v2);

		// Count unique values
		int uniqueCount = 1;
		for (int i=1; i<v2.length; i++)
			if (v2[i] != v2[i-1])
				uniqueCount ++;

		// Extract the unique values
		double[][] u = new double[2][uniqueCount];
		double lastValue = v2[0];
		int start = 0;
		int end = 0;
		int index = 0;

		for (int i=1; i<v2.length; i++)
		{
			if (v2[i] == lastValue)
				end = i;
			else
			{
				u[0][index] = lastValue;
				u[1][index] = (start + end) / 2.0;
				lastValue = v2[i];
				start = i;
				end = i;
				index ++;
			}
		}

		u[0][index] = lastValue;
		u[1][index] = (start + end) / 2;

		return u;
	}

	/**
	 *	Get the ranks [0..length-1] of the values in a vector. If a
	 *	value appears multiple times, its rank will be the middle
	 *	location of all appearences.
	 *	@param		v			The values
	 *	@return					The ranks
	 */
	public static double[] getRanks(double[] v)
	{
		double[][] u = getUniqueValues(v);
		double[] r = new double[v.length];
		for (int i=0; i<v.length; i++)
			r[i] = u[1][Arrays.binarySearch(u[0], v[i])];
		return r;
	}

	/**
	 *	Get the normal scores of a set of ranks.
	 *	@param		r			The ranks
	 *	@return					The normal scores
	 */
	public static double[] getNormalScores(double[] r)
	{
		NormalDistribution nd = new NormalDistribution();
		double[] z = new double[r.length];
		for (int i=0; i<r.length; i++)
			z[i] = nd.inverse((r[i] + 0.5) / (r.length + 1));
		return z;
	}

	/**
	 *	Compute the Pearson correlation between two vectors.
	 *	@param		v1			The first vector
	 *	@param		v2			The second vector
	 *	@return					The correlation
	 */
	public static double getPearsonCorrelation(double[] v1, double[] v2)
	{
		if (v1.length != v2.length)
			throw new RuntimeException("The input vectors have different lengths.");

		long count = 0;
		double sum1 = 0;
		double sum2 = 0;
		double sumSq1 = 0;
		double sumSq2 = 0;
		double sum12 = 0;

		for (int i=0; i<v1.length; i++)
		{
			count ++;
			sum1 += v1[i];
			sum2 += v2[i];
			sumSq1 += v1[i] * v1[i];
			sumSq2 += v2[i] * v2[i];
			sum12 += v1[i] * v2[i];
		}

		return (count * sum12 - sum1 * sum2) /
		       Math.pow(count * sumSq1 - sum1 * sum1, 0.5) / Math.pow(count * sumSq2 - sum2 * sum2, 0.5);
	}

	/**
	 *	Compute the Spearman correlation between two vectors.
	 *	@param		v1			The first vector
	 *	@param		v2			The second vector
	 *	@return					The correlation
	 */
	public static double getSpearmanCorrelation(double[] v1, double[] v2)
	{
		if (v1.length != v2.length)
			throw new RuntimeException("The input vectors have different lengths.");
		double[] r1 = getRanks(v1);
		double[] r2 = getRanks(v2);
		return getPearsonCorrelation(r1, r2);
	}

	/**
	 *	Compute the normal score correlation between two vectors.
	 *	@param		v1			The first vector
	 *	@param		v2			The second vector
	 *	@return					The correlation
	 */
	public static double getNormalScoreCorrelation(double[] v1, double[] v2)
	{
		if (v1.length != v2.length)
			throw new RuntimeException("The input vectors have different lengths.");
		double[] r1 = getRanks(v1);
		double[] z1 = getNormalScores(r1);
		r1 = null;
		double[] r2 = getRanks(v2);
		double[] z2 = getNormalScores(r2);
		r2 = null;
		return getPearsonCorrelation(z1, z2);
	}
}