package org.gersteinlab.act.data;

import java.io.*;
import java.util.*;

/**
 *	<pre>
 *	This class defines an interval of genomic locations. Positions are
 *	0-based, and the intervals are left-closed, right-open. In other words,
 *	if a sequence contains n locations, the whole interval is represented as
 *	[0, n).
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (May 7, 2010)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class Interval implements Comparable<Interval>, Serializable
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	/**
	 *	The official symbol for the plus strand.
	 */
	public static final char PLUS_STRAND = '+';

	/**
	 *	The official symbol for the minus strand.
	 */
	public static final char MINUS_STRAND = '-';

	/**
	 *	The official symbol to indicate that the strand is not
	 *	important.
	 */
	public static final char ANY_STRAND = '.';


	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	/**
	 *	The chromosome of the interval.
	 */
	String chr = null;

	/**
	 *	The starting position of the interval, 0-based, included.
	 */
	protected long start = -1;

	/**
	 *	The ending position of the interval, 0-based, not included.
	 */
	protected long end = -1;

	/**
	 *	The strand of the interval.
	 */
	protected char strand = '.';

	/**
	 *	The double value of the interval.
	 */
	protected double doubleValue = Double.NaN;

	/**
	 *	Additional attributes of the interval.
	 */
	protected Map<String, String> attributes = null;


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create a dummy interval.
	 */
	public Interval()
	{
	}

	/**
	 *	Create an interval.
	 *	@param		chrIn			Chromosome of the interval
	 *	@param		startIn			The starting position of the interval, 0-based, included
	 *	@param		endIn			The ending position of the interval, 0-based, not included
	 */
	public Interval(String chrIn, long startIn, long endIn)
	{
		chr = chrIn;
		start = startIn;
		end = endIn;
	}

	/**
	 *	Create an interval.
	 *	@param		chrIn			Chromosome of the interval
	 *	@param		startIn			The starting position of the interval, 0-based, included
	 *	@param		endIn			The ending position of the interval, 0-based, not included
	 *	@param		strandIn		The strand
	 */
	public Interval(String chrIn, long startIn, long endIn, char strandIn)
	{
		chr = chrIn;
		start = startIn;
		end = endIn;
		if (strandIn == PLUS_STRAND || strandIn == MINUS_STRAND || strandIn == ANY_STRAND)
			strand = strandIn;
		else
			throw new RuntimeException("Unknown strand: " + strandIn);
	}

	/**
	 *	Get the chromosome of the interval.
	 *	@return					The chromosome
	 */
	public String getChr()
	{
		return chr;
	}

	/**
	 *	Get the starting position of the interval.
	 *	@return					The starting position of the interval, 0-based, included
	 */
	public long getStart()
	{
		return start;
	}

	/**
	 *	Get the ending position of the interval.
	 *	@return					The ending position of the interval, 0-based, not included
	 */
	public long getEnd()
	{
		return end;
	}

	/**
	 *	Get the strand of the interval.
	 *	@return					The strand
	 */
	public long getStrand()
	{
		return strand;
	}

	/**
	 *	Get the double value of the interval.
	 *	@return					The value
	 */
	public double getDoubleValue()
	{
		return doubleValue;
	}

	/**
	 *	Get the value of an attribute.
	 *	@param		attrName		Name of the attribute
	 *	@return					The value, null if not available
	 */
	public String getAttribute(String attrName)
	{
		if (attributes == null)
			return null;
		else
			return attributes.get(attrName);
	}

	/**
	 *	Set the chromosome of the interval.
	 *	@param		chrIn			The chromosome of the interval
	 */
	public void setChr(String chrIn)
	{
		chr = chrIn;
	}

	/**
	 *	Set the starting position of the interval.
	 *	@param		startIn			The starting position of the interval, 0-based, included
	 */
	public void setStart(long startIn)
	{
		start = startIn;
	}

	/**
	 *	Set the ending position of the interval.
	 *	@param		endIn			The ending position of the interval, 0-based, not included
	 */
	public void setEnd(long endIn)
	{
		end = endIn;
	}

	/**
	 *	Set the strand of the interval.
	 *	@param		strandIn		The strand
	 */
	public void setStrand(char strandIn)
	{
		if (strandIn == PLUS_STRAND || strandIn == MINUS_STRAND || strandIn == ANY_STRAND)
			strand = strandIn;
		else
			throw new RuntimeException("Unknown strand: " + strandIn);
	}

	/**
	 *	Set the double value of the interval.
	 *	@param		doubleValueIn		The value
	 */
	public void setDoubleValue(double doubleValueIn)
	{
		doubleValue = doubleValueIn;
	}

	/**
	 *	Set the value of an attribute. If there was an old value, it
	 *	will be overriden.
	 *	@param		attrName		Name of the attribute
	 *	@param		attrValue		Value of the attribute
	 */
	public void setAttribute(String attrName, String attrValue)
	{
		if (attributes == null)
			attributes = new LinkedHashMap<String, String>();
		attributes.put(attrName, attrValue);
	}

	/**
	 *	Get the length of the interval.
	 *	@return					The length
	 */
	public long getLength()
	{
		return end - start;
	}

	/**
	 *	Compare this interval with an input interval based on natural
	 *	ordering.
	 *	@param		intervalIn		The input interval
	 *	@return					-1 if this interval should be ordered first<br />
	 *						1 if this interval should be ordered second<br />
	 *						0 if the two intervals have the same order
	 */
	public int compareTo(Interval intervalIn)
	{
		return forwardCompareTo(intervalIn);
	}

	/**
	 *	Compare this interval with an input interval based on forward
	 *	ordering.
	 *	@param		intervalIn		The input interval
	 *	@return					-1 if this interval should be ordered first<br />
	 *						1 if this interval should be ordered second<br />
	 *						0 if the two intervals have the same order
	 */
	public int forwardCompareTo(Interval intervalIn)
	{
		     if (chr.compareTo(intervalIn.chr) < 0)		return -1;
		else if (chr.compareTo(intervalIn.chr) > 0)		return 1;
		else if (start		< intervalIn.start)		return -1;
		else if (start		> intervalIn.start)		return 1;
		else if (end		< intervalIn.end)		return -1;
		else if (end		> intervalIn.end)		return 1;
		else if (strand		< intervalIn.strand)		return -1;
		else if (strand		> intervalIn.strand)		return 1;
		else							return 0;
	}

	/**
	 *	Compare this interval with an input interval based on backward
	 *	ordering.
	 *	@param		intervalIn		The input interval
	 *	@return					-1 if this interval should be ordered first<br />
	 *						1 if this interval should be ordered second<br />
	 *						0 if the two intervals have the same order
	 */
	public int backwardCompareTo(Interval intervalIn)
	{
		     if (chr.compareTo(intervalIn.chr) < 0)		return -1;
		else if (chr.compareTo(intervalIn.chr) > 0)		return 1;
		else if (end		> intervalIn.end)		return -1;	// Note these four lines
		else if (end		< intervalIn.end)		return 1;
		else if (start		> intervalIn.start)		return -1;
		else if (start		< intervalIn.start)		return 1;
		else if (strand		< intervalIn.strand)		return -1;
		else if (strand		> intervalIn.strand)		return 1;
		else							return 0;
	}

	/**
	 *	Check if this interval is semantically equivalent to an input
	 *	object.
	 *	@param		o			The input object
	 *	@return					Whether the objects are semantically equivalent
	 */
	public boolean equals(Object o)
	{
		if (!(o instanceof Interval))
			return false;
		Interval intervalIn = (Interval)o;
		     if (chr.equals(intervalIn.chr) == false)		return false;
		else if (start		!= intervalIn.start)		return false;
		else if (end		!= intervalIn.end)		return false;
		else if (strand		!= intervalIn.strand)		return false;
		else							return true;
	}

	/**
	 *	Get the hash code of this interval.
	 *	@return					The hash code
	 */
	public int hashCode()
	{
		return toString().hashCode();
	}

	/**
	 *	Check if this interval intersects with an input interval.
	 *	@param		intervalIn		The input interval
	 *	@return					Whether the intervals intersect
	 */
	public boolean intersects(Interval intervalIn)
	{
		     if (this.chr.equals(intervalIn.chr) == false)
			return false;
		else if ((this.strand == PLUS_STRAND && intervalIn.strand == MINUS_STRAND) ||
		         (this.strand == MINUS_STRAND && intervalIn.strand == PLUS_STRAND))
			return false;
		else if (this.end <= intervalIn.start ||
		         this.start >= intervalIn.end)
			return false;
		else
			return true;
	}

	/**
	 *	Get a string representation of the interval.
	 *	@return					The string representation
	 */
	public String toString()
	{
		return chr + (strand == ANY_STRAND ?"" :(""+strand)) + ":[" + start + "," + end + ")";
	}
}
