package org.gersteinlab.act.io;

import org.gersteinlab.act.data.*;
import org.gersteinlab.act.util.*;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

/**
 *	<pre>
 *	This class defines a reader of data files.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (May 7, 2010)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public abstract class DataFilesReader
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	/**
	 *	The input reader.
	 */
	protected BufferedReader br = null;


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create a new reader with source data from an input reader.
	 *	@param		rIn			The input reader
	 */
	public DataFilesReader(Reader rIn)
	{
		br = new BufferedReader(rIn);
	}

	/**
	 *	Get a data file reader for a data file. This method tries to
	 *	identify the file format based on the file extension.
	 *	@param		config			Configurations
	 *	@param		dataFile		The data file
	 *	@return					The data file reader
	 */
	public static DataFilesReader getReader(Config config, File dataFile)
	throws ClassNotFoundException,
	       FileNotFoundException,
	       IllegalAccessException,
	       InstantiationException,
	       InvocationTargetException,
	       NoSuchMethodException
	{
		String fileName = dataFile.getName();
		int sep = fileName.lastIndexOf(".");
		if (sep == -1 || sep == fileName.length() - 1)
			throw new RuntimeException("Unable to identify the format of the data file " + dataFile.getName());
		String fileExt = fileName.substring(sep + 1).toLowerCase();
		String readerClassName = config.getConfig("class.reader." + fileExt);
		if (readerClassName == null)
			throw new RuntimeException("Unknown format of the data file " + dataFile.getName());
		else
			return (DataFilesReader)(Class.forName(readerClassName).getConstructor(Reader.class).newInstance(new FileReader(dataFile)));
	}

	/**
	 *	Reset the reader.
	 */
	public void reset() throws IOException
	{
		br.reset();
	}

	/**
	 *	Close the reader.
	 */
	public void close() throws IOException
	{
		br.close();
	}

	/**
	 *	Read the next interval from the reader.
	 *	@return					The next interval<br />
	 *						null if not more
	 */
	public abstract Interval readInterval() throws IOException;

	/**
	 *	Read all intervals into a collection.
	 *	@param		output			Collection for storing the intervals
	 */
	public void readIntervals(Collection<Interval> output) throws IOException
	{
		Interval interval = readInterval();
		while (interval != null)
		{
			output.add(interval);
			interval = readInterval();
		}
	}

	/**
	 *	Read the intervals that contain a given value into a collection.
	 *	@param		output			Collection for storing the intervals
	 *	@param		value			The value
	 */
	public void readIntervals(Collection<Interval> output, double value) throws IOException
	{
		Interval interval = readInterval();
		while (interval != null)
		{
			if (interval.getDoubleValue() == value)
				output.add(interval);
			interval = readInterval();
		}
	}

	/**
	 *	Get the total number of intervals.
	 *	@return					The total number
	 */
	public int countIntervals() throws IOException
	{
		int count = 0;
		Interval interval = readInterval();
		while (interval != null)
		{
			count ++;
			interval = readInterval();
		}
		return count;
	}
}