package org.gersteinlab.act.io;

import org.gersteinlab.act.data.*;
import org.gersteinlab.act.util.*;

import java.io.*;
import java.util.*;

/**
 *	<pre>
 *	This class defines a reader of data files in GTF format. See
 *	http://genome.ucsc.edu/FAQ/FAQformat.html for definition.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (May 7, 2010)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class GtfFilesReader extends DataFilesReader
{
	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create a new reader with source data from an input reader.
	 *	@param		rIn			The input reader
	 */
	public GtfFilesReader(Reader rIn)
	{
		super(rIn);
	}

	/**
	 *	Read the next interval from the reader.
	 */
	public Interval readInterval() throws IOException
	{
		// Read a line from the input reader
		String line = br.readLine();
		Interval interval = null;

		while (true)
		{
			// Case 1: no more unread lines
			if (line == null)
				break;

			// Case 2: an unused line
			else if (line.length() == 0 || line.startsWith("track") || line.startsWith("#"))
			{
				line = br.readLine();
				continue;
			}

			// Case 3: a data line
			else
			{
				// Parse the fields
				StringTokenizer2 st = new StringTokenizer2(line, "\t", false);
				String chr = st.nextToken();
				String source = st.nextToken();
				String feature = st.nextToken();
				long start = Long.parseLong(st.nextToken()) - 1;
				long end = Long.parseLong(st.nextToken());
				String valueStr = st.nextToken();
				double value = valueStr.equals(".") ?Double.NaN :Double.parseDouble(valueStr);
				char strand = st.nextToken().charAt(0);
				String frame = st.nextToken();

				// Create the interval
				interval = new Interval(chr, start, end, strand);
				interval.setDoubleValue(value);
				interval.setAttribute("source",		source);
				interval.setAttribute("feature",	feature);
				interval.setAttribute("frame",		frame);

				// Store additional attributes
				if (st.hasMoreTokens())
				{
					StringTokenizer2 st2 = new StringTokenizer2(st.nextToken(), ";", false);
					while (st2.hasMoreTokens())
					{
						String attribute = st2.nextToken().trim();
						if (attribute.length() == 0)
							break;
						int sepPt = attribute.indexOf(" ");
						String attrName = attribute.substring(0, sepPt - 1);
						String attrValue = attribute.substring(sepPt + 1);
						interval.setAttribute(attrName, attrValue);
					}
				}

				// Store comments
				if (st.hasMoreTokens())
					interval.setAttribute("comments", st.nextToken());

				// Create and return the interval
				break;
			}
		}
		return interval;
	}
}