package org.gersteinlab.act.io;

import org.gersteinlab.act.data.*;

import java.io.*;
import java.util.*;

/**
 *	<pre>
 *	This class defines a reader of data files in WIG format. See
 *	http://genome.ucsc.edu/FAQ/FAQformat.html for definition.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (May 7, 2010)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class WigFilesReader extends DataFilesReader
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	/**
	 *	Fixed step format
	 */
	public static final int STEP_FIXED	= 1;

	/**
	 *	Variable step format
	 */
	public static final int STEP_VARIABLE	= 2;


	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	/**
	 *	Segment of the last interval
	 */
	protected String seg = null;

	/**
	 *	Starting position of the last interval, 0-based, included
	 */
	protected long intervalStart = -1;

	/**
	 *	Double value of the last interval
	 */
	protected double doubleValue = Double.NaN;

	/**
	 *	The step size of the intervals
	 */
	protected int step = -1;

	/**
	 *	The span of each interval
	 */
	protected int span = -1;

	/**
	 *	The step type of intervals
	 */
	protected int stepType = -1;


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create a new reader with source data from an input reader.
	 *	@param		rIn			The input reader
	 */
	public WigFilesReader(Reader rIn)
	{
		super(rIn);
	}

	/**
	 *	Read the next interval from the reader.
	 */
	public Interval readInterval() throws IOException
	{
		// Read a line from the input reader
		String line = br.readLine();
		Interval interval = null;

		while (true)
		{
			// Case 1: no more unread lines
			if (line == null)
				break;

			// Case 2: an unused line
			else if (line.startsWith("track"))
				doubleValue = Double.NaN;

			// Case 3a: a header line for fixed step format
			else if (line.startsWith("fixedStep"))
			{
				StringTokenizer st = new StringTokenizer(line, " \t", false);
				st.nextToken();		// Skip fixedStep
				String segStr = st.nextToken();		// Treated as chromosome
				seg = segStr.substring(segStr.indexOf("=") + 1);
				String startStr = st.nextToken();
				intervalStart = Long.parseLong(startStr.substring(startStr.indexOf("=") + 1)) - 1;
				String stepStr = st.nextToken();
				step = Integer.parseInt(stepStr.substring(stepStr.indexOf("=") + 1));
				if (st.hasMoreTokens())
				{
					String spanStr = st.nextToken();
					span = Integer.parseInt(spanStr.substring(spanStr.indexOf("=") + 1));
				}
				else
					span = 1;	// Default span

				stepType = STEP_FIXED;
				doubleValue = Double.NaN;
			}

			// Case 3b: a header line for variable step format
			else if (line.startsWith("variableStep"))
			{
				StringTokenizer st = new StringTokenizer(line, " \t", false);
				st.nextToken();		// Skip variableStep
				String segStr = st.nextToken();
				seg = segStr.substring(segStr.indexOf("=") + 1);
				if (st.hasMoreTokens())
				{
					String spanStr = st.nextToken();
					span = Integer.parseInt(spanStr.substring(spanStr.indexOf("=") + 1));
				}
				else
					span = 1;	// Default span

				stepType = STEP_VARIABLE;
				doubleValue = Double.NaN;
			}

			// Case 4: a data line
			else
			{
				if (stepType == STEP_FIXED)
				{
					if (!Double.isNaN(doubleValue))	// Not the first value in this track
						intervalStart += step;
					doubleValue = Double.parseDouble(line);
				}
				else if (stepType == STEP_VARIABLE)
				{
					StringTokenizer st = new StringTokenizer(line, " \t", false);
					intervalStart = Long.parseLong(st.nextToken()) - 1;
					doubleValue = Double.parseDouble(st.nextToken());
				}
				long intervalEnd = intervalStart + span;
				interval = new Interval(seg, intervalStart, intervalEnd);
				interval.setDoubleValue(doubleValue);
				break;
			}
			line = br.readLine();
		}
		return interval;
	}
}