package org.gersteinlab.act.util;

import java.io.*;
import java.util.*;

/**
 *	<pre>
 *	This class defines a set of configurations. It is similar to
 *	java.util.Properties, but the order of the configurations is maintained.
 *	Also, if the same name is shared by multiple entries, all the entries
 *	will be stored and associated with the name.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (May 7, 2010)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class Config
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	/**
	 *	The name to values mapping of the configurations.
	 */
	Map<String, List<String>> configMap = new LinkedHashMap<String, List<String>>();;


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create a new configuration set by reading from a configuration
	 *	file.
	 *	@param		configFile			The configuration file
	 */
	public Config(File configFile) throws IOException
	{
		BufferedReader br = new BufferedReader(new FileReader(configFile));
		String line = br.readLine();
		while (line != null)
		{
			line = line.trim();
			if (!line.startsWith("#") && !line.startsWith("!"))
			{
				int sep = line.indexOf("=");
				if (sep != -1)
				{
					String configName = line.substring(0, sep).trim();
					String configValue = (sep == line.length() - 1) ?"" :line.substring(sep + 1).trim();
					List<String> configValues = configMap.get(configName);
					if (configValues == null)
					{
						configValues = new ArrayList<String>();
						configMap.put(configName, configValues);
					}
					configValues.add(configValue);
				}
			}
			line = br.readLine();
		}
		br.close();
	}

	/**
	 *	Get the values of a certain configuration.
	 *	@param		configName			The name of the configuration
	 *	@return						The values,<br />
	 *							null if there is no such entry
	 */
	public List<String> getConfigs(String configName)
	{
		return configMap.get(configName);
	}

	/**
	 *	Get the value a certain configuration. If there are multiple
	 *	values, the first one will be returned.
	 *	@param		configName			The name of the configuration
	 *	@return						The value,<br />
	 *							null if there is no such entry
	 */
	public String getConfig(String configName)
	{
		List<String> configValues = configMap.get(configName);
		return (configValues == null) ?null :configValues.get(0);
	}

	/**
	 *	Get the names of the configurations.
	 *	@return						The names
	 */
	public Set<String> getConfigNames()
	{
		return configMap.keySet();
	}
}
